// Copyright (C) 2005, 2007 International Business Machines and others.
// All Rights Reserved.
// This code is published under the Common Public License.
//
// $Id: genome_ipopt_main.cpp,v 1.6 2010/12/15 08:09:54 andrones Exp $
//
// Authors:  Carl Laird, Andreas Waechter     IBM    2005-08-10

#include "IpIpoptApplication.hpp"
#include "genome_ipopt_nlp.hpp"
#include <cstring>
#include <unistd.h>

// for printf
#ifdef HAVE_CSTDIO
# include <cstdio>
#else
# ifdef HAVE_STDIO_H
#  include <stdio.h>
# else
#  error "don't have header file for stdio"
# endif
#endif

using namespace Ipopt;
using namespace std;


char interactions_filename[1000]="";
//interactions_chr3_freq2plusS5_S5andU5.txt";
//interactions_chr3_freq15plusS2_S2andU2.txt";
//interactions_chr3_freq3plusS6_S6andU6.txt
//interactions_chr3_freq2plusS5_S5andU5.txt
//interactions_chr3_freq20plusU1.txt
//interactions_chr3_chr4_inter20plus_intraall.txt
//interactions_chr4_freq20plusU1.txt
//3interaction_chr3.txt 
//interactions_chr3_strongest.txt   
char secondary_inter_filename[1000]=""; // Diament & Tuller, May 2014
int bp_per_locus = 1000;
double sphere_radius=1.0;
// 0.0111 is the length of 1kpb if the radius is 1 (a magic number from 2 papers)
// if the radius is 10, length is 0.111
//double max_dist = 0.0111 * bp_per_locus/1000;

double max_dist = 0.0091;   // 0.0091 microm max length of a (packed around nucleosomes) 1000bp
double min_dist = 0.0066;   // 0.0066 microm min length of a (packed around nucleosomes) 1000bp
//double min_clash_dist = 0.002;  // 0.002 microm distance
double min_clash_dist = 0.03;   // 30 nm distance
double min_clash_dist_inter = 0.1;
int add_rDNA = 1;


char output_pdb[1000] = "output.pdb";
char input_txt[1000] = "";
//char initial_pdb[] = "initial.pdb";
// char initial_interaction_matrix[] = "initial_int_mat.txt";
// char optimal_interaction_matrix[] = "optimal_int_mat.txt";
// char interaction_achievement_matrix[] = "inter_achiev_mat.txt";
char output_binary_diffusion[1000] = ""; 
int diffusion_operation_type = 0;

char output_cplex_input[1000] = "";     // output for us, input for cplex 
char input_cplex_output[1000] = "";     // input for us, output for cplex
 
int num_chroms = 16; 
int chromosome[16]; // 16 chromosomes in yeast
// this array will have 1 for the chromosomes I'm interested in, and 0 for the ones I'm not interested in

char chrom_string[100] = "1";
int use_weights = 0;
int weight_of_inter = 1;
char rDNA_interactions_filename[1000] = "";
char rDNA_frequency_normalizer = 1;

char unseen_interactions_filename[1000] = "";
double weight_unseen = 1; 

void usage (const char *prog)
// PRE:  None
// POST: Prints the usage message and exits
{
    printf ("\nUsage: %s -i <interactions_file> [options]\n\n", prog);
    printf ("Options:\n");
    printf ("  -c <chromosomes>, e.g. 1 2 3 1,3 (not supporting - right now)\n");
    printf ("  -r <resolution>, default %d\n", bp_per_locus);
    printf ("  -s <sphere_radius>, default %g\n", sphere_radius);
    printf ("  -d <distance_perkb>, default %g\n", max_dist);
    printf ("  -o <output_pdb>\n");
    printf ("  -w <0|1>, default %d\n\tHow to use weights:\n",  use_weights);
    printf ("\t\t0  don't use any weights\n");
    printf ("\t\t1  weight = ligation frequency in the objective function\n");
    printf ("\t\t2  use ligation frequencies as constraints\n");
    printf ("\t\t3  use -1/log(pvalues) as weights in the objective function\n");
    printf ("\t\t4  use -1/log(qvalues) as weights in the objective function\n");
    printf ("\t\t5  use a function that transforms frequency into wish distance, in the objective; freq are summed\n");
    printf ("\t\t6  input are wish distance; dist are averaged\n"); // Diament & Tuller, May 2014
    printf ("  -y <weight of inter/intra large>, puts this weight on the inter and intra between centromeric arms (applies only if w is 5 or 6). The local intra will have weight 1\n");
    printf ("  -b <binary_output_file_for_diffusion>, default none\n");
    printf ("\tWrite a binary matrix and exit\n");
    printf ("  -a <diffusion-operation-type>, 0=binary, 1=freq; 2=counts\n");
    printf ("  -p <cplex_input_file>, default none\n");
    printf ("\tWrite the problem in cplex format .lp and exit\n");
    printf ("  -t <cplex_output_file>, default none\n");
    printf ("\tRead the .sol file, write pdb file and exit. A pdb filename must be given.\n");
    printf ("  -n <rDNA_interaction_file>, default none\n");
    printf ("  -e <rDNA_frequency_normalizer>. Divide the rDNA frequency by this number. Default %d\n", rDNA_frequency_normalizer);
    printf ("  -f <file_unseen_interactions>, default none. Applies when w is 5 or 6.\n");
    printf ("  -g <weight_unseen_interactions>, for example 0.1. Applies when w is 5 or 6.\n");
    printf ("  -x <txt_file>. Takes as input the txt file with coordinates and outputs the PDB file. It still needs the -i, -c and -r options, but it doesn't run the optimization\n");     
    printf ("  -h <secondary interactions file>\n\n"); // Diament & Tuller, May 2014
    exit (0);
}


void get_arguments (int argc, char *argv[])
// PRE:  None
// POST: Get the parameters from the command line
{
    int c;
    extern char *optarg;
    extern int optind;
    int errflag = 0;

    while ((c = getopt (argc, argv, "a:b:c:d:e:f:g:h:i:j:n:r:s:w:y:o:p:t:x:?")) != -1)
    {
        switch (c)
        {
            case 'c':
                strcpy (chrom_string, optarg);
                break;
            case 'n':
                strcpy (rDNA_interactions_filename, optarg);
                break;    
            case 'e':
                rDNA_frequency_normalizer = atoi (optarg);
                break;
            case 'f':
                strcpy (unseen_interactions_filename, optarg);
                break;    
            case 'g':
                weight_unseen = atof (optarg);
                break;                
            case 'i':
                strcpy (interactions_filename, optarg);
                break;
            case 'h': // Diament & Tuller, May 2014
            	strcpy (secondary_inter_filename, optarg);
            	break;
            case 'r':
                bp_per_locus = atoi (optarg);
                break;
            case 's':
                sphere_radius = atof (optarg);
                break;
            case 'j':
                min_clash_dist_inter = atof (optarg);
                break;                
            case 'd':
                max_dist = atof (optarg);
                break;
            case 'o':
                strcpy (output_pdb, optarg);
                break;
            case 'w':
                use_weights = atoi (optarg);
                break;      
            case 'y':
                weight_of_inter = atoi(optarg);
                break;    
            case 'b':
                strcpy (output_binary_diffusion, optarg);
                break;
            case 'a':
                diffusion_operation_type = atoi (optarg);
                break;
            case 'p':
                strcpy (output_cplex_input, optarg);
                break;
            case 't':
                strcpy (input_cplex_output, optarg);
                break;    
            case 'x':
                strcpy (input_txt, optarg);
                break;                
            case '?':
            default:
                errflag = 1;
        }
    }
    if (errflag || optind != argc || argc < 2)
        usage (argv[0]);

}


int main(int argc, char* argv[])
{

    get_arguments (argc, argv);

    // figure out which chromosomes I want
    for (int i=0; i < num_chroms; i++)
    {
        chromosome[i] = 0;
    }
    char *token;
    token = strtok(chrom_string,",");
    while (token != NULL)
    {
        int index = atoi (token) - 1;
        //printf ("token: %s\n", token);
        chromosome[index] = 1;
        token = strtok(NULL,",");
    } 

    // the max_dist is per kb, so adjust it so that it depends on the resolution    
    max_dist = max_dist * bp_per_locus/1000.0;
    // same for min_dist
    min_dist = min_dist * bp_per_locus/1000.0;
    //double min_interaction_dist = 0.01;   // not USED

  // Create a new instance of your nlp
  //  (use a SmartPtr, not raw)  
    //printf ("Making genome object\n");
    GENOME *mygenome = new GENOME(bp_per_locus, min_dist, max_dist,
    		interactions_filename, chromosome, add_rDNA,
    		rDNA_interactions_filename, use_weights,
    		unseen_interactions_filename, secondary_inter_filename); // Diament & Tuller, May 2014
    //printf ("Genome object done\n");
    
    if (strcmp (input_txt, "") != 0)
    {
        printf ("Reading the txt input\n");
        mygenome->read_txt_input(input_txt);
        //mygenome->print_pdb_genome(output_pdb);
        mygenome->print_1D_3D_genome(output_pdb);
        return 1;
    }
    else if (strcmp (output_binary_diffusion, "") != 0)
    {
        mygenome->save_adjacency_matrix_for_diffusion (output_binary_diffusion, diffusion_operation_type);    
        return 1;
    }
    else if (strcmp (output_cplex_input, "") != 0)
    // just write the .lp file for cplex
    {
        SmartPtr<genome_ipopt_nlp> mynlp = new genome_ipopt_nlp (mygenome, min_clash_dist, min_clash_dist_inter, output_pdb, sphere_radius, use_weights, bp_per_locus, rDNA_frequency_normalizer, weight_of_inter, weight_unseen);    
        mynlp->write_cplex_input (output_cplex_input);
        return 1;
    }
    else if (strcmp (input_cplex_output, "") != 0)
    // read the file in .sol format (cplex output) and write a pdb file
    {
        mygenome->read_cplex_output(input_cplex_output);
        mygenome->print_pdb_genome(output_pdb);
        return 1;
    }
  //mygenome->save_interaction_adjacency_matrix (initial_interaction_matrix);
  //mygenome->print_pdb_genome (initial_pdb);
  
  SmartPtr<TNLP> mynlp = new genome_ipopt_nlp (mygenome, min_clash_dist, min_clash_dist_inter, output_pdb, sphere_radius, use_weights, bp_per_locus, rDNA_frequency_normalizer, weight_of_inter, weight_unseen);      

  // Create a new instance of IpoptApplication
  //  (use a SmartPtr, not raw)
  SmartPtr<IpoptApplication> app = new IpoptApplication();

  // Change some options
  // Note: The following choices are only examples, they might not be
  //       suitable for your optimization problem.
  app->Options()->SetNumericValue("tol", 1e-2);
  app->Options()->SetStringValue("mu_strategy", "adaptive");
  app->Options()->SetStringValue("output_file", "ipopt.out");
  app->Options()->SetStringValue("hessian_approximation", "limited-memory");  
  //app->Options()->SetNumericValue("max_iter", 500.0);
  // The following overwrites the default name (ipopt.opt) of the
  // options file
  // app->Options()->SetStringValue("option_file_name", "hs071.opt");

  // Intialize the IpoptApplication and process the options
  ApplicationReturnStatus status;
  status = app->Initialize();
  if (status != Solve_Succeeded) {
    printf("\n\n*** Error during initialization!\n");
    return (int) status;
  }
  
  // Ask Ipopt to solve the problem
  status = app->OptimizeTNLP(mynlp);
   
  //mygenome->save_interaction_achievement (interaction_achievement_matrix);
  //mygenome->save_optimal_distances (optimal_interaction_matrix);   

  if (status == Solve_Succeeded) {
    printf("\n\n*** The problem solved!\n");
  }
  else {
    printf("\n\n*** The problem FAILED!\n");
  }

  // As the SmartPtrs go out of scope, the reference count
  // will be decremented and the objects will automatically
  // be deleted.

  return 1;     //(int) status;
  
}
